#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

This is a library for parsing deb lines into deb822-format data.
"""
# pylint: disable=missing-function-docstring, missing-class-docstring
# These aren't really relevant for unit testing (which is mostly automatic.)

import unittest

from urllib.error import URLError

from . import ppa
from . import util

class PPATestCase(unittest.TestCase):
    def setUp(self):
        self.source = ppa.PPALine(
            'ppa:system76/pop',
            fetch_data=False
        )

    def test_uri(self):
        self.assertEqual(
            self.source.uris,
            ['http://ppa.launchpad.net/system76/pop/ubuntu']
        )

    def test_name(self):
        self.assertEqual(self.source.name, 'ppa-system76-pop')

    def test_suite(self):
        self.assertEqual(self.source.suites, [util.DISTRO_CODENAME])

    def test_components(self):
        self.assertEqual(self.source.components, ['main'])

    def test_type(self):
        self.assertEqual(self.source.types, [util.AptSourceType.BINARY])

    def test_options(self):
        self.assertFalse(self.source.options)

    def test_internet_features(self):
        try:
            self.source.load_from_ppa()
            self.assertEqual(self.source.name, 'Pop!_OS PPA')
            self.assertNotEqual(self.source.ppa_info, {})
        except util.RepoError as err:
            # pylint: disable=no-else-raise, no-member
            # This is a valid time to else raise since we're catching
            # a different type of error.
            if isinstance(err.args[-1], URLError):
                if "Name or service not known" in str(err.args[-1].reason):
                    raise unittest.SkipTest(
                        'Can\'t test internet features without network.'
                    )
                elif "Connection refused" in str(err.args[-1].reason):
                    raise unittest.SkipTest(
                        'Couldn\'t fetch details from network.'
                    )
            else:
                raise err



#!/usr/bin/python3

"""
Copyright (c) 2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

Module for getting an argparser. Used by apt-manage
"""

import argparse
import inspect

from repolib import command as cmd

from .. import __version__

def get_argparser():
    """ Get an argument parser with our arguments.

    Returns:
        An argparse.ArgumentParser.
    """
    parser = argparse.ArgumentParser(
        prog='apt-manage',
        description='Manage software sources and repositories',
        epilog='apt-manage version: {}'.format(__version__.__version__)
    )

    parser.add_argument(
        '-b',
        '--debug',
        action='count',
        help=argparse.SUPPRESS
    )

    subparsers = parser.add_subparsers(
        help='...',
        dest='action',
        metavar='COMMAND'
    )

    subcommands = []
    for i in inspect.getmembers(cmd, inspect.isclass):
        obj = getattr(cmd, i[0])
        subcommands.append(obj)

    for i in subcommands:
        i.init_options(subparsers)

    return parser



"""
Copyright (c) 2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

Module for modifying repos in CLI applications.
"""

from argparse import SUPPRESS
import sys

from . import command

from ..legacy_deb import LegacyDebSource
from ..source import Source, SourceError
from ..util import get_source_path
from ..system import SystemSource

class Modify(command.Command):
    """ Modify subcommand.

    The modify command will allow making modifications to the specified repository.
    It requires providing a repository to modify.

    Options:
        --enable, -e
        --disable, -d
        --default-mirror
        --name
        --add-suite
        --remove-suite
        --add-component
        --remove-component
        --add-uri
        --remove-uri
        --add-option
        --remove-option
    """

    @classmethod
    def init_options(cls, subparsers):
        """ Sets up this command's options parser."""

        parser = subparsers.add_parser(
            'modify',
            help='Change a configured repository.'
        )
        parser.add_argument(
            'repository',
            nargs='*',
            default=['system'],
            help='The repository to modify. Default is the system repository.'
        )

        modify_enable = parser.add_mutually_exclusive_group(
            required=False
        )
        modify_enable.add_argument(
            '-e',
            '--enable',
            action='store_true',
            help='Enable the repository, if disabled.'
        )
        modify_enable.add_argument(
            '-d',
            '--disable',
            action='store_true',
            help=(
                'Disable the repository, if enabled. The system repository cannot '
                'be disabled.'
            )
        )
        modify_enable.add_argument(
            '--default-mirror',
            help=SUPPRESS
            #help='Sets the default mirror for the system source.'
        )

        # Name
        parser.add_argument(
            '-n',
            '--name',
            help='Set the repository name to NAME'
        )

        # Suites
        parser.add_argument(
            '--add-suite',
            metavar='SUITE[,SUITE]',
            help=(
                'Add the specified suite(s) to the repository. Multiple '
                'repositories should be separated with commas. NOTE: Legacy deb '
                'repositories may only contain one suite.'
            )
        )
        parser.add_argument(
            '--remove-suite',
            metavar='SUITE[,SUITE]',
            help=(
                'Remove the specified suite(s) from the repository. Multiple '
                'repositories should be separated with commas. NOTE: The last '
                'suite in a repository cannot be removed.'
            )
        )

        # Components
        parser.add_argument(
            '--add-component',
            metavar='COMPONENT[,COMPONENT]',
            help=(
                'Add the specified component(s) to the repository. Multiple '
                'repositories should be separated with commas.'
            )
        )
        parser.add_argument(
            '--remove-component',
            metavar='COMPONENT[,COMPONENT]',
            help=(
                'Remove the specified component(s) from the repository. Multiple '
                'repositories should be separated with commas. NOTE: The last '
                'component in a repository cannot be removed.'
            )
        )

        # URIs
        parser.add_argument(
            '--add-uri',
            metavar='URI[,URI]',
            help=(
                'Add the specified URI(s) to the repository. Multiple '
                'repositories should be separated with commas. NOTE: Legacy deb '
                'repositories may only contain one uri.'
            )
        )
        parser.add_argument(
            '--remove-uri',
            metavar='URI[,URI]',
            help=(
                'Remove the specified URI(s) from the repository. Multiple '
                'repositories should be separated with commas. NOTE: The last '
                'uri in a repository cannot be removed.'
            )
        )

        # Options
        parser.add_argument(
            '--add-option',
            metavar='OPTION=VALUE[,OPTION=VALUE]',
            help=SUPPRESS
            # help=(
            #     'Add the specified option(s) and value(s) to the repository. '
            #     'Multiple option-value pairs should be separated with commas.'
            # )
        )
        parser.add_argument(
            '--remove-option',
            metavar='OPTION=VALUE[,OPTION=VALUE]',
            help=SUPPRESS
            # help=(
            #     'Remove the specified option(s) and value(s) from the repository. '
            #     'Multiple option-value pairs should be separated with commas.'
            # )
        )

    def __init__(self, log, args, parser):
        super().__init__(log, args, parser)
        self.source = None
        self.count = 0

    def finalize_options(self, args):
        """ Finish setting up our options/arguments."""
        super().finalize_options(args)
        self.repo = ' '.join(args.repository)
        self.enable = args.enable
        self.disable = args.disable

        self.actions = {}

        self.actions['endisable'] = None
        for i in ['enable', 'disable']:
            if getattr(args, i):
                self.actions['endisable'] = i

        for i in [
                'default_mirror',
                'name',
                'add_uri',
                'remove_uri',
                'add_suite',
                'remove_suite',
                'add_component',
                'remove_component',
                'add_option',
                'remove_option'
        ]:
            self.actions[i] = getattr(args, i)

    def run(self):
        """ Run the command."""
        self.log.debug('Modifying repository %s', self.repo)
        full_path = get_source_path(self.repo, log=self.log)

        if not full_path:
            self.log.error('Could not find source %s', self.repo)
            return False

        if self.repo == 'system':
            self.source = SystemSource()
        elif full_path.suffix == '.sources':
            self.source = Source(ident=self.repo)
        else:
            self.source = LegacyDebSource(ident=self.repo)

        self.source.load_from_file()
        self.log.debug('Actions taken: \n%s', self.actions)
        self.log.info('Source before:\n%s', self.source.make_source_string())

        for i in self.actions:
            getattr(self, i)(self.actions[i])

        self.log.info('Source after:\n%s', self.source.make_source_string())

        if self.count == 0:
            self.log.error('No changes specified, no actions taken')
            return False

        if not self.debug:
            self.source.save_to_disk()
            return True

        return True

    def default_mirror(self, value):
        """ Set the default mirror if this is the system source."""
        if not value:
            # No value provided, take no action
            return
        self.count += 1

        if self.repo == 'system':
            self.source.default_mirror = value

    def name(self, value):
        """ Set the source name. """
        if not value:
            # No value provided, take no action
            return
        self.count += 1
        self.log.info('Set name: %s', value)

        self.source.name = value

    def endisable(self, value):
        """ Enable or disable the source. """
        if not value:
            # No value provided, take no action
            return
        self.count += 1
        self.log.info('Enable/Disable: %s', value)

        if value == 'disable':
            self.source.enabled = False
            return
        self.source.enabled = True

    def add_uri(self, value):
        """ Add URIs to the source. """
        if not value:
            # No value provided, take no action
            return
        self.count += 1
        self.log.info('Add uris: %s', value)

        for uri in value.split():
            if uri not in self.source.uris:
                try:
                    self.source.uris = self.source.uris + [uri]
                except SourceError:
                    self.log.error('The URI "%s" is malformed', uri)
                    sys.exit(1)
                self.log.debug('Added uri: %s', uri)
    def remove_uri(self, value):
        """ Remove URIs from the source. """
        if not value:
            # No value provided, take no action
            return
        self.count += 1
        self.log.info('Remove uris: %s', value)

        for uri in value.split():
            if uri in self.source.uris:
                uris = self.source.uris
                uris.remove(uri)
                self.source.uris = uris
                self.log.debug('Removed uri %s', uri)
    def add_suite(self, value):
        """ Add suites to the source. """
        if not value:
            # No value provided, take no action
            return
        self.count += 1
        self.log.info('Add suites: %s', value)

        for suite in value.split():
            if suite not in self.source.suites:
                self.source.suites = self.source.suites + [suite]
                self.log.debug('Added suite: %s', suite)

    def remove_suite(self, value):
        """ Remove suites from the source. """
        if not value:
            # No value provided, take no action
            return
        self.count += 1
        self.log.info('Remove suites: %s', value)

        for suite in value.split():
            if suite in self.source.suites:
                suites = self.source.suites
                suites.remove(suite)
                self.source.suites = suites
                self.log.debug('Removed suite %s', suite)

    def add_component(self, value):
        """ Add components to the source. """
        if not value:
            # No value provided, take no action
            return
        self.count += 1
        self.log.info('Add components: %s', value)

        for component in value.split():
            if component not in self.source.components:
                self.source.components = self.source.components + [component]
                self.log.debug('Added component: %s', component)

    def remove_component(self, value):
        """ Remove components from the source. """
        if not value:
            # No value provided, take no action
            return
        self.count += 1
        self.log.info('Remove components: %s', value)

        for component in value.split():
            if component in self.source.components:
                components = self.source.components
                components.remove(component)
                self.source.components = components
                self.log.debug('Removed component %s', component)

    def add_option(self, value):
        """ Add options to the source. """
        if not value:
            # No value provided, take no action
            return
        self.count += 1
        self.log.info('Add options: %s', value)

        for option in value.split():
            if option not in self.source.options:
                self.source.options = self.source.options + [option]
                self.log.debug('Added option: %s', option)

    def remove_option(self, value):
        """ Remove options from the source. """
        if not value:
            # No value provided, take no action
            return
        self.count += 1
        self.log.info('Remove options: %s', value)

        for option in value.split():
            if option in self.source.options:
                options = self.source.options
                options.remove(option)
                self.source.options = options
                self.log.debug('Removed option %s', option)



#!/usr/bin/python3

"""
Copyright (c) 2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

Module for adding repos to the system in CLI applications.
"""

from ..deb import DebLine
from ..legacy_deb import LegacyDebSource
from ..ppa import PPALine
from ..util import DISTRO_CODENAME, CLEAN_CHARS

from . import command

class Add(command.Command):
    """ Add subcommand.

    The add command is used for adding new software sources to the system. It
    requires root.

    Options:
        --disable, -d
        --source-code, -s
        --expand, -e
    """

    @classmethod
    def init_options(cls, subparsers):
        """ Sets up this command's options parser.

        Returns:
            The subparser for this command.
        """
        options = subparsers.add_parser(
            'add',
            help='Add a new repository to the system.'
        )
        options.add_argument(
            'deb_line',
            nargs='*',
            default=['822styledeb'],
            help='The deb line of the repository to add'
        )
        options.add_argument(
            '-d',
            '--disable',
            action='store_true',
            help='Add the repository and then set it to disabled.'
        )
        options.add_argument(
            '-s',
            '--source-code',
            action='store_true',
            help='Also enable source code packages for the repository.'
        )
        options.add_argument(
            '-e',
            '--expand',
            action='store_true',
            help='Display expanded details about the repository before adding it.'
        )
        options.add_argument(
            '-n',
            '--name',
            default=['x-repolib-default-name'],
            help='A name to set for the new repo'
        )
        options.add_argument(
            '-i',
            '--identifier',
            default=['x-repolib-default-id'],
            help='The identifier/filename to use for the new repo'
        )

    # pylint: disable=too-few-public-methods
    # Thinking of ways to add more, but otherwise this is just simple.

    def __init__(self, log, args, parser):
        super().__init__(log, args, parser)

        self.verbose = False
        if self.args.debug > 1:
            self.verbose = True

        self.expand = args.expand
        self.source_code = args.source_code
        self.disable = args.disable
        try:
            name = args.name.split()
        except AttributeError:
            name = args.name
        try:
            ident = args.identifier.split()
        except AttributeError:
            ident = args.identifier
        self.name = ' '.join(name)
        self.ident = '-'.join(ident).translate(CLEAN_CHARS)

    def set_names(self, source):
        """Set up names for the source.

        Arguments:
            source(repolib.Source): The source for which to set names
        """
        source.make_names()

        if self.name != 'x-repolib-default-name':
            self.log.debug('Got Name: %s', self.name)
            source.name = self.name

        if self.ident != 'x-repolib-default-id':
            self.log.debug('Got Ident: %s', self.ident)
            source.ident = self.ident.lower()

    def run(self):
        """ Run the command."""
        # pylint: disable=too-many-branches
        # We just need all these different checks.

        debline = ' '.join(self.args.deb_line)
        if self.args.deb_line == '822styledeb':
            self.parser.print_usage()
            self.log.error('A repository is required.')
            return False

        if debline.startswith('http') and len(debline.split()) == 1:
            debline = f'deb {debline} {DISTRO_CODENAME} main'

        new_source = LegacyDebSource()
        if debline.startswith('ppa:'):
            add_source = PPALine(debline, verbose=self.verbose)

        elif debline.startswith('deb'):
            self.expand = False
            add_source = DebLine(debline)

        else:
            self.log.critical(
                'The line "%s" is malformed. Double-check the spelling.',
                debline
            )
            return False

        new_source.name = add_source.name
        new_source.sources.append(add_source)

        if not debline.startswith('deb-src'):
            src_source = add_source.copy()
            src_source.enabled = False
        else:
            src_source = add_source

        src_source.enabled = self.source_code

        if not debline.startswith('deb-src'):
            new_source.sources.append(src_source)

        new_source.load_from_sources()

        add_source.enabled = True

        if self.disable:
            for repo in new_source.sources:
                repo.enabled = False
            new_source.enabled = False

        self.set_names(new_source)

        self.log.debug(new_source.name)
        self.log.debug(new_source.filename)

        if self.args.debug > 0:
            self.log.info('Debug mode set, not saving.')
            for src in new_source.sources:
                print(f'{src.dump()}\n')
            self.log.info('Filename to save: %s', new_source.filename)
            print(f'{new_source.make_deblines()}')

        if self.expand:
            print(new_source.sources[0].make_source_string())
            print(f'{add_source.ppa_info["description"]}\n')
            print('Press [ENTER] to contine or Ctrl + C to cancel.')
            input()

        if self.args.debug == 0:
            new_source.save_to_disk()
            return True

        return False



#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.
"""

from enum import Enum
from pathlib import Path
from urllib.parse import urlparse

SOURCES_DIR = '/etc/apt/sources.list.d'
TESTING = False

class RepoError(Exception):
    """ Exception from this module."""

    def __init__(self, *args, code=1, **kwargs):
        """Exception with a source object

        Arguments:
            code (:obj:`int`, optional, default=1): Exception error code.
    """
        super().__init__(*args, **kwargs)
        self.code = code

try:
    import distro
    DISTRO_CODENAME = distro.codename()
except ImportError:
    DISTRO_CODENAME = 'linux'

class AptSourceType(Enum):
    """ Helper Enum to simplify saving data. """
    BINARY = "deb"
    SOURCE = "deb-src"

class AptSourceEnabled(Enum):
    """ Helper Enum to translate between bool data and the Deb822 format. """
    TRUE = 'yes'
    FALSE = 'no'

    def get_bool(self):
        """ Return a bool based on the value. """
        # pylint: disable=comparison-with-callable
        # This doesnt seem to actually be a callable in this case.
        if self.value == "yes":
            return True

        return False

CLEAN_CHARS = {
    33: None,
    64: 45,
    35: 45,
    36: 45,
    37: 45,
    94: 45,
    38: 45,
    42: 45,
    41: None,
    40: None,
    43: 45,
    61: 45,
    91: None,
    92: None,
    93: None,
    123: None,
    125: None,
    124: 95,
    63: None,
    47: 95,
    46: 45,
    60: 95,
    62: 95,
    44: 95,
    96: None,
    126: None,
    32: 95,
    58: None,
    59: None,
}

def url_validator(url):
    """ Validate a url and tell if it's good or not.

    Arguments:
        url (str): The URL to validate.

    Returns:
        `True` if `url` is not malformed, otherwise `False`.
    """
    try:
        # pylint: disable=no-else-return,bare-except
        # A) We want to return false if the URL doesn't contain those parts
        # B) We need this to not throw any exceptions, regardless what they are
        result = urlparse(url)
        if not result.scheme:
            return False
        if result.scheme == 'x-repolib-name':
            return False
        if result.netloc:
            # We need at least a scheme and a netlocation/hostname or...
            return all([result.scheme, result.netloc])
        elif result.path:
            # ...a scheme and a path (this allows file:/// URIs which are valid)
            return all([result.scheme, result.path])
        return False
    except:
        return False

def get_source_path(name, log=None):
    """ Tries to get the full path to the source.

    This is necessary because some sources end in .list, others in .sources

    Returns:
        pathlib.Path for the actual full path.
    """
    full_name = f'{name}.sources'
    full_path = get_sources_dir() / full_name
    if log:
        log.debug('Trying to load %s', full_path)
    if full_path.exists():
        if log:
            log.debug('Path %s exists!', full_path)
        return full_path

    full_name = f'{name}.list'
    full_path = get_sources_dir() / full_name
    if log:
        log.debug('Trying to load %s', full_path)
    if full_path.exists():
        if log:
            log.debug('Path %s exists!', full_path)
        return full_path
    return None

def get_sources_dir(testing=False):
    """ Get the path to the sources dir.

    Returns:
        pathlib.Path: The Sources dir.
    """
    # pylint: disable=global-statement
    # We want to stop using the old dir and use the testing dir on subsequent
    # calls.
    if testing:
        global SOURCES_DIR
        SOURCES_DIR = '/tmp/repolib_testing'
    # pylint: enable=global-statement
    sources_dir = Path(SOURCES_DIR)
    sources_dir.mkdir(parents=True, exist_ok=True)
    return sources_dir

# pylint: disable=inconsistent-return-statements
# This is a better way to check these
def validate_debline(valid):
    """ Basic checks to see if a given debline is valid or not.

    Arguments:
        valid (str): The line to validate.

    Returns:
        True if the line is valid, False otherwise.
    """
    if valid.startswith('#'):
        valid = valid.replace('#', '')
        valid = valid.strip()

    if valid.startswith("deb"):
        words = valid.split()
        for word in words:
            if url_validator(word):
                return True

    elif valid.startswith("ppa:"):
        if "/" in valid:
            return True

    else:
        if valid.endswith('.flatpakrepo'):
            return False
        return url_validator(valid)



#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=missing-function-docstring, missing-class-docstring
# These aren't really relevant for unit testing (which is mostly automatic.)

import unittest

from . import legacy_deb
from . import util


class LegacyTestCase(unittest.TestCase):

    def setUp(self):
        with open(util.get_sources_dir(testing=True) / 'test.list', mode='w') as test_file:
            test_file.write(
                'deb [arch=amd64] https://example.com ubuntu main universe\n'
                'deb-src [arch=amd64] https://example.com ubuntu main universe\n'
            )
        with open(util.get_sources_dir(testing=True) / 'test2.list', mode='w') as test_file:
            test_file.write(
                '## Added/managed by repolib ##\n'
                '#\n'
                '## X-Repolib-Name: example-com\n'
                'deb [arch=armel,amd64 lang=en_US] http://example.com ubuntu main\n'
                'deb-src [arch=armel,amd64 lang=en_US] http://example.com ubuntu main\n'
            )
        with open(util.get_sources_dir(testing=True) / 'test3.list', mode='w') as test_file:
            test_file.write(
                '## Added/managed by repolib ##\n'
                '#\n'
                '## X-Repolib-Name: test no suites\n'
                'deb [trusted=yes] file:///var/cache/apt/archives ./\n'
                'deb-src [trusted=yes] file:///var/cache/apt/archives ./\n'
            )
        self.source = legacy_deb.LegacyDebSource(ident='test')
        self.source.load_from_file()

    def test_no_suites(self):
        opts = {'Trusted': 'yes'}
        types = [util.AptSourceType.BINARY, util.AptSourceType.SOURCE]
        source = legacy_deb.LegacyDebSource(ident='test3')
        source.load_from_file()

        self.assertEqual(source.ident, 'test3')
        self.assertEqual(source.name, 'test no suites')
        self.assertEqual(source.types, types)
        self.assertEqual(source.uris, ['file:///var/cache/apt/archives'])
        self.assertEqual(source.suites, ['./'])
        self.assertFalse(source.components)
        self.assertDictEqual(source.options, opts)

    def test_load_from_file(self):
        opts = {'Architectures': 'amd64'}
        self.assertEqual(self.source.filename, 'test.list')
        self.assertEqual(len(self.source.sources), 2)
        for debsource in self.source.sources:
            self.assertEqual(len(debsource.types), 1)
            self.assertEqual(debsource.uris, ['https://example.com'])
            self.assertDictEqual(debsource.options, opts)
            self.assertEqual(debsource.suites, ['ubuntu'])
            self.assertEqual(debsource.components, ['main', 'universe'])

    def test_make_deblines(self):
        # pylint: disable=line-too-long
        # unittest doesn't like when this is split to a sane length. We need to
        # have it on one ugly line so that the assertion is Equal.
        expected = '## Added/managed by repolib ##\n#\n## X-Repolib-Name: example-com\ndeb [arch=amd64] https://example.com ubuntu main universe\ndeb-src [arch=amd64] https://example.com ubuntu main universe\n'
        actual = self.source.make_deblines()
        self.assertEqual(actual, expected)

    def test_validate(self):
        testline1 = '# deb http://example.com ubuntu main'
        testline2 = 'deb http://example.com ubuntu main'
        testline3 = '#deb-src http://example.com ubuntu main'
        testline4 = 'deb-src http://example.com ubuntu main'
        testline5 = '# dbe http://example.com ubuntu main'
        testline6 = 'dba http://example.com ubuntu main'
        testline7 = 'deb file:///var/cache/apt/archives ./'
        testline8 = 'deb [trusted=yes] file:///var/cache/apt/archives ./'

        for line in [testline1, testline2, testline3, testline4, testline7, testline8]:
            self.assertTrue(util.validate_debline(line))

        for line in [testline5, testline6]:
            self.assertFalse(util.validate_debline(line))

    def test_save_to_disk(self):
        with open(util.get_sources_dir(testing=True) / 'test2.list', mode='r') as expected_file:
            expected_data = expected_file.readlines()

        with open(util.get_sources_dir(testing=True) / 'test.list', mode='r') as source_file:
            source_data = source_file.readlines()
        self.assertNotEqual(source_data, expected_data)

        self.source.uris = ['http://example.com']
        self.source.options = {
            'Architectures': 'armel amd64',
            'Languages': 'en_US'
        }
        self.source.components = ['main']
        self.source.filename = 'test3.list'
        self.source.save_to_disk()

        with open(util.get_sources_dir(testing=True) / 'test3.list', mode='r') as source_file:
            source_file.seek(0)
            source_data = source_file.readlines()
        self.assertEqual(source_data, expected_data)



#!/usr/bin/python3

"""
Copyright (c) 2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

Module for listing repos on the system in CLI applications.
"""

import traceback

from ..deb import DebLine
from ..legacy_deb import LegacyDebSource
from ..source import Source
from ..util import get_sources_dir, RepoError
from .. import get_all_sources

from . import command

class List(command.Command):
    """ List subcommand

    The list command lists available software sources as well as details about
    sources. With no further options, it lists all configured sources. With a
    configured source, it lists details about the specified source.

    Options:
        --legacy, -l
        --verbose, v
    """

    @classmethod
    def init_options(cls, subparsers):
        """ Sets up this command's options parser.

        Returns:
            The subparser for this command.
        """
        options = subparsers.add_parser(
            'list',
            help=(
                'List configured repositories. If a repository name is provided, '
                'show details about that repository.'
                )
        )

        options.add_argument(
            'repository',
            nargs='*',
            default=['x-repolib-all-sources'],
            help='The repository to list details about.'
        )
        options.add_argument(
            '-v',
            '--verbose',
            action='store_true',
            help='Display details of all configured repositories.'
        )
        options.add_argument(
            '-l',
            '--legacy',
            action='store_true',
            help='Include listing details about entries configured in sources.list'
        )
        options.add_argument(
            '-n',
            '--no-names',
            action='store_true',
            dest='names',
            help='Do not print names of repositories.'
        )



    def __init__(self, log, args, parser):
        super().__init__(log, args, parser)
        self.verbose = args.verbose
        self.legacy = args.legacy
        self.no_names = args.names
        self.source = ' '.join(args.repository)
        self.sources_dir = get_sources_dir()

    def list_all_sources(self):
        """ List all sources on the system, potentially with their info."""
        # pylint: disable=too-many-branches
        # The branches involved control program output. Maybe this can be
        # split into separate methods.
        try:
            sources_list_d_file = self.sources_dir.parent / 'sources.list'
        except FileNotFoundError:
            sources_list_d_file = None

        if not self.no_names:
            print('Configured sources:')

        sources, errors = get_all_sources(get_system=True, get_exceptions=True)

        for source in sources:
            self.log.debug('Found source file %s', source.filename)
            if self.no_names:
                print(f'{source.ident}')
            else:
                print(f'{source.ident} - {source.name}')
            if self.verbose:
                print(f'{source.make_source_string()}')

        if sources_list_d_file and self.legacy:
            print('\n Legacy sources.list entries:\n')
            with sources_list_d_file.open(mode='r') as file:
                for line in file:
                    if "cdrom:" in line:
                        line = ''
                    try:
                        source = DebLine(line)
                        print(f'{source.make_debline()}')
                    except RepoError:
                        pass
        if errors:
            print('\nThe following files have formatting errors:')
            for err in errors:
                print(err)
            if self.verbose or self.debug:
                print('\nDetails for the failing files:')
                for err in errors:
                    print(f'{err}:')
                    with open(err) as error_file:
                        print(error_file.read())
                    print('Stack Trace:')
                    traceback.print_tb(errors[err].__traceback__)
                    print('\n')


        return True

    def get_source_path(self):
        """ Tries to get the full path to the source.

        This is necessary because some sources end in .list, others in .sources

        Returns:
            source.Source for the actual full path.
        """
        full_name = f'{self.source}.sources'
        full_path = self.sources_dir / full_name
        self.log.debug('Trying to load %s', full_path)
        if full_path.exists():
            self.log.debug('Path %s exists!', full_path)
            source = Source(filename=full_path.name)
            source.load_from_file()
            return source

        full_name = f'{self.source}.list'
        full_path = self.sources_dir / full_name
        self.log.debug('Trying to load %s', full_path)
        if full_path.exists():
            self.log.debug('Path %s exists!', full_path)
            leg = LegacyDebSource(filename=full_path.name)
            leg.load_from_file()
            return leg

        raise RepoError('The path does not exist (checked .sources and .list files.')

    def run(self):
        """ Run the command. """
        ret = False
        if self.source == 'x-repolib-all-sources':
            ret = self.list_all_sources()

        else:
            try:
                source = self.get_source_path()
            except RepoError:
                self.log.error(
                    "Couldn't find the source file for %s. Check your spelling.",
                    self.source
                )
                return False

            print(f'Details for source {self.source}:\n{source.make_source_string()}')
            ret = True
        return ret



#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.
"""

from .source import Source
from .system import SystemSource
from .legacy_deb import LegacyDebSource
from .deb import DebLine
from .ppa import PPALine
from .util import AptSourceEnabled, AptSourceType, RepoError
from . import util
from . import ppa
from . import __version__

VERSION = __version__.__version__

# pylint: disable=broad-except
# We want to be broad in catching exceptions here, as failure could mean
# applications unexpectedly close
def get_all_sources(get_system=False, get_exceptions=False):
    """ Returns a list of all the sources on the system.

    Arguments:
        get_system (bool): Whether to include the system repository or not.
        get_exceptions (bool): Whether to return information about failures.

    Returns:
        Without `get_exceptions`, return the :obj:`list` of :obj:`Source`
        With `get_exceptions`, return: (
            :obj:`list` of :obj:`Source`,
            :obj:`dict` of :obj:`Exception`
        )
    """
    sources_path = util.get_sources_dir()
    sources_files = sources_path.glob('*.sources')
    list_files = sources_path.glob('*.list')

    sources = []
    errors = {}

    if get_system:
        source = SystemSource()
        sources.append(source)

    for file in sources_files:
        if file.stem == 'system':
            continue
        source = Source(filename=file.name)
        try:
            source.load_from_file()
        except Exception as err:
            errors[file.stem] = err
        else:
            # The source should not be listed if it is empty
            has_uris = len(source.uris) > 0
            has_suites = len(source.suites) > 0
            if has_uris and has_suites:
                sources.append(source)

    for file in list_files:
        source = LegacyDebSource(filename=file.name)
        try:
            source.load_from_file()
        except Exception as err:
            errors[file] = err
        else:
            # The source should not be listed if it is empty
            has_uris = len(source.uris) > 0
            has_suites = len(source.suites) > 0
            if has_uris and has_suites:
                sources.append(source)

    if get_exceptions:
        return sources, errors
    return sources



#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

This is a library for parsing deb lines into deb822-format data.
"""
# pylint: disable=missing-function-docstring, missing-class-docstring
# These aren't really relevant for unit testing (which is mostly automatic.)

import unittest

from . import system
from . import util

class SystemTestCase(unittest.TestCase):
    #pylint: disable=invalid-name
    # We're inheriting from unittest in this case.
    sources_dir = util.get_sources_dir(testing=True)

    def setUpModule(self):
        with open(self.sources_dir / 'modified_system.sources', mode='w') as sources_file:
            sources_file.write(
                'X-Repolib_Name: Pop!_OS Sources\n'
                'Enabled: no\n'
                'Types: deb\n'
                'URIs: http://apt.pop-os.org/ubuntu http://archive.ubuntu.com/ubuntu\n'
                'Suites: focal focal-updates focal-test\n'
                'Components: main universe multiverse test\n'
            )

    def setUp(self):
        with open(self.sources_dir / 'system.sources', mode='w') as sources_file:
            sources_file.write(
                'X-Repolib_Name: Pop!_OS Sources\n'
                'Enabled: no\n'
                'Types: deb deb-src\n'
                'URIs: http://apt.pop-os.org/ubuntu\n'
                'Suites: focal focal-updates focal-backports\n'
                'Components: main universe multiverse restricted\n'
            )
        self.source = system.SystemSource()

    def test_load_system_source(self):
        self.assertEqual(self.source.filename, 'system.sources')
        self.assertNotEqual(self.source.uris, [])
        self.assertNotEqual(self.source.suites, [])
        self.assertNotEqual(self.source.components, [])
        self.assertNotEqual(self.source.types, [])

    def test_set_component_enabled(self):
        expected = ['main', 'universe', 'multiverse', 'test']
        self.source.set_component_enabled(component='test')
        self.source.set_component_enabled(component='restricted', enabled=False)

        self.assertEqual(self.source.components, expected)

    def test_set_suite_enabled(self):
        expected = ['focal', 'focal-updates', 'focal-test']
        self.source.set_suite_enabled(suite='focal-backports', enabled=False)
        self.source.set_suite_enabled(suite='focal-test')

        self.assertEqual(self.source.suites, expected)

    def test_set_source_enabled(self):
        expected = [util.AptSourceType.BINARY]
        self.source.set_source_enabled(False)

        self.assertEqual(self.source.types, expected)



#! /usr/bin/python3

"""
repolib PPA Signing Key Support

From software-properties

Copyright (c) 2004-2009 Canonical Ltd.
          (c) 2019-2020 Ian Santopietro <isantop@gmail.com>

Author: Michael Vogt <mvo@debian.org>
        Ian Santopietro <isantop@gmail.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA
"""
#pylint: disable=too-many-ancestors
# If we want to use the subclass, we don't have a lot of options.

from __future__ import print_function

import json
import subprocess
import time
import urllib.parse
import urllib.request
from http.client import HTTPException
from urllib.error import HTTPError, URLError

from . import source, util

DISTRO_CODENAME = util.DISTRO_CODENAME

SKS_KEYSERVER = 'https://keyserver.ubuntu.com/pks/lookup?op=get&options=mr&exact=on&search=0x%s'
# maintained until 2015
LAUNCHPAD_PPA_API = 'https://launchpad.net/api/devel/%s/+archive/%s'
LAUNCHPAD_USER_API = 'https://launchpad.net/api/1.0/%s'
LAUNCHPAD_USER_PPAS_API = 'https://launchpad.net/api/1.0/%s/ppas'
LAUNCHPAD_DISTRIBUTION_API = 'https://launchpad.net/api/1.0/%s'
LAUNCHPAD_DISTRIBUTION_SERIES_API = 'https://launchpad.net/api/1.0/%s/%s'
# Specify to use the system default SSL store; change to a different path
# to test with custom certificates.
LAUNCHPAD_PPA_CERT = "/etc/ssl/certs/ca-certificates.crt"

class PPALine(source.Source):
    """ A source specifically for Launchpad PPAs

    These are typically given in the format ppa:owner/name. This is a
    convenience function which makes adding these sources easier. Most of this
    code was adopted from SoftwareProperties.

    Arguments:
        line (str): The ppa: format line.
        fetch_data (bool): Whether to try and fetch metadata from LP.
    """
    # pylint: disable=too-many-instance-attributes
    # These just have more data than a normal source, and most of these are
    # @properties anyway (due to the inheritance from source.Source).

    def __init__(self, line, fetch_data=True, verbose=False):
        super().__init__()
        self.ppa_line = line
        self.verbose = verbose
        if not self.ppa_line.startswith('ppa:'):
            raise util.RepoError("The PPA %s is malformed!" % self.ppa_line)

        self.load_from_ppa(fetch_data=fetch_data)

    def load_from_ppa(self, fetch_data=True):
        """ Load PPA information from the PPA.

        Arguments:
            fetch_data (bool): Whether to fetch metadata from Launchpad.
        """
        self.init_values()
        self.enabled = 'yes'

        raw_ppa = self.ppa_line.replace('ppa:', '').split('/')
        ppa_owner = raw_ppa[0]
        ppa_name = raw_ppa[1]

        ppa_info = self.ppa_line.split(":")
        ppa_uri = 'http://ppa.launchpad.net/{}/ubuntu'.format(ppa_info[1])
        self.set_source_enabled(False)
        self.uris = [ppa_uri]
        self.suites = [DISTRO_CODENAME]
        self.components = ['main']
        ppa_name = ppa_info[1].split('/')
        self.name = 'ppa-{}'.format('-'.join(ppa_name))
        self.ident = '{}'.format(self.name)
        if fetch_data:
            self.ppa_info = get_info_from_lp(ppa_owner, ppa_name[1])
            if self.verbose:
                print(self.ppa_info)
            self.name = self.ppa_info['displayname']
        self.enabled = util.AptSourceEnabled.TRUE

    #pylint: disable=arguments-differ
    # Just doing something different than the super class.
    def make_name(self):
        """ Make a name suitable for a PPA.

        Returns:
            str: The name generated.
        """
        name = self.ppa_line.replace(':', '-')
        name = name.replace('/', '-')

        return f'{name}'

    def save_to_disk(self, save=True):
        """
        Saves the PPA to disk, and fetches the signing key.
        """
        self._get_ppa_key()
        if save:
            super().save_to_disk()

    def copy(self, source_code=True):
        """ Copies the source and returns an identical source object.

        Arguments:
            source_code (bool): if True, output an identical source, except with
                source code enabled.

        Returns:
            A Source() object identical to self.
        """
        new_source = PPALine(self.ppa_line)
        new_source = self._copy(new_source, source_code=source_code)
        return new_source

    def _get_ppa_key(self):
        if self.ppa_line:
            add_key(self.ppa_info['signing_key_fingerprint'])

def get_info_from_lp(owner_name, ppa):
    """ Attempt to get information on a PPA from launchpad over the internet.

    Arguments:
        owner_name (str): The Launchpad user owning the PPA.
        ppa (str): The name of the PPA

    Returns:
        json: The PPA information as a JSON object.
    """
    if owner_name[0] != '~':
        owner_name = '~' + owner_name
    lp_url = LAUNCHPAD_PPA_API % (owner_name, ppa)
    data = _get_https_content_py3(lp_url, True)
    return json.loads(data)

def _get_https_content_py3(lp_url, accept_json, retry_delays=None):
    if retry_delays is None:
        retry_delays = []

    trynum = 0
    err = None
    sleep_waits = iter(retry_delays)
    headers = {"Accept": "application/json"} if accept_json else {}

    while True:
        trynum += 1
        try:
            request = urllib.request.Request(str(lp_url), headers=headers)
            lp_page = urllib.request.urlopen(request)
            return lp_page.read().decode("utf-8", "strict")
        except (HTTPException, URLError) as errr:
            err = util.RepoError(
                "Error reading %s (%d tries): %s" % (lp_url, trynum, errr.reason),
                errr)
            # do not retry on 404. HTTPError is a subclass of URLError

            # pylint: disable=no-member
            # Yes it does, and this works fine. Not sure why things complain.
            if isinstance(errr, HTTPError) and errr.code == 404:
                break
        try:
            time.sleep(next(sleep_waits))
        except StopIteration:
            break

    raise err

def add_key(fingerprint):
    """ Add a key for a PPA into the system configuration.

    Arguments:
        fingerprint (str): The fingerprint of the key to add.
    """
    apt_key_cmd = "apt-key adv --keyserver keyserver.ubuntu.com --recv-keys".split()
    # apt_key_cmd.append(ppa_info['signing_key_fingerprint'])
    apt_key_cmd.append(fingerprint)
    subprocess.run(
        apt_key_cmd,
        check=False
    )



#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.
"""
#pylint: disable=too-many-ancestors
# If we want to use the subclass, we don't have a lot of options.

import re

import dbus
from debian import deb822

from . import util


class SourceError(Exception):
    """ Exception from a source object."""

    def __init__(self, *args, code=1, **kwargs):
        """Exception with a source object

        Arguments:
            code (:obj:`int`, optional, default=1): Exception error code.
    """
        super().__init__(*args, **kwargs)
        self.code = code

class Source(deb822.Deb822):
    """ A Deb822 object representing a software source.

    Provides a dict-like interface for accessing and modifying DEB822-format
    sources, as well as options for loading and saving them from disk.
    """
    # pylint: disable=too-many-instance-attributes
    # We want to provide easy access to these data

    options_d = {
        'arch': 'Architectures',
        'lang': 'Languages',
        'target': 'Targets',
        'pdiffs': 'PDiffs',
        'by-hash': 'By-Hash',
        'allow-insecure': 'Allow-Insecure',
        'allow-weak': 'Allow-Weak',
        'allow-downgrade-to-insecure': 'Allow-Downgrade-To-Insecure',
        'trusted': 'Trusted',
        'signed-by': 'Signed-By',
        'check-valid-until': 'Check-Valid-Until',
        'valid-until-min': 'Valid-Until-Min',
        'valid-until-max': 'Valid-Until-Max'
    }

    outoptions_d = {
        'Architectures': 'arch',
        'Languages': 'lang',
        'Targets': 'target',
        'PDiffs': 'pdiffs',
        'By-Hash': 'by-hash',
        'Allow-Insecure': 'allow-insecure',
        'Allow-Weak': 'allow-weak',
        'Allow-Downgrade-To-Insecure': 'allow-downgrade-to-insecure',
        'Trusted': 'trusted',
        'Signed-By': 'signed-by',
        'Check-Valid-Until': 'check-valid-until',
        'Valid-Until-Min': 'valid-until-min',
        'Valid-Until-Max': 'valid-until-max'
    }
    options_re = re.compile(r'[^@.+]\[([^[]+.+)\]\ ')
    uri_re = re.compile(r'\w+:(\/?\/?)[^\s]+')

    def __init__(self, *args, filename=None, ident=None, **kwargs):
        super().__init__(*args, **kwargs)
        if filename:
            self.filename = filename
        if ident:
            self.ident = ident

    def load_from_file(self, filename=None, ident=None):
        """ Loads the data from a file path.

        Arguments:
            filename (str): The name of the file on disk.
        """
        self._set_filename_ident(filename, ident)
        if not self.ident:
            raise SourceError("No filename to load from")

        full_path = util.get_sources_dir() / self.filename

        with open(full_path, mode='r') as source_file:
            super().__init__(source_file)

    def save_to_disk(self, save=True):
        """ Saves the source to disk."""
        if not self.filename:
            raise SourceError('No filename to save to specified')
        full_path = util.get_sources_dir() / self.filename

        if save:
            try:
                with open(full_path, mode='w') as sources_file:
                    sources_file.write(self.dump())
            except PermissionError:
                bus = dbus.SystemBus()
                privileged_object = bus.get_object('org.pop_os.repolib', '/Repo')
                privileged_object.output_file_to_disk(self.filename, self.dump())
                privileged_object.exit()

    def make_source_string(self):
        """ Makes a printable string of the source.

        This method is intended to provide output in a user-friendly format. For
        output representative of the actual data, use dump() instead.

        Returns:
            A str which can be printed to console.
        """
        # The source can't be enabled if there are no uris or suites
        no_uris = len(self.uris) <= 0
        no_suites = len(self.suites) <= 0
        if no_uris or no_suites:
            self.enabled = False

        if not self.name:
            self.name = self.filename.replace('.sources', '')

        toprint = self.dump()
        toprint = toprint.replace('X-Repolib-Name', 'Name')
        return toprint

    def set_source_enabled(self, enabled):
        """ Convenience method to set a source with source_code enabled.

        If source code is enabled, then the Types for self will be both
        BINARY and SOURCE. Otherwise it will be just BINARY.

        Arguments:
            enabled(bool): Wether or not to enable source-code.
        """
        if enabled:
            self.types = [util.AptSourceType.BINARY, util.AptSourceType.SOURCE]
        else:
            self.types = [util.AptSourceType.BINARY]

    # pylint: disable=arguments-differ
    # We're fundamentally doing a different thing than our super class.
    def copy(self, source_code=True):
        """ Copies the source and returns an identical source object.

        Arguments:
            source_code (bool): if True, output an identical source, except with
                source code enabled.

        Returns:
            A Source() object identical to self.
        """
        new_source = Source()
        new_source = self._copy(new_source, source_code=source_code)
        return new_source

    def make_name(self, prefix=''):
        """ Create a name for this source. """
        if len(self.uris) > 0:
            uri = self.uris[0].replace('/', ' ')
            uri_list = uri.split()
            name = '{}{}'.format(
                prefix,
                '-'.join(uri_list[1:]).translate(util.CLEAN_CHARS)
            )
        else:
            # Use the ident as a fallback as it should be good enough
            name = self.ident

        return name

    def init_values(self):
        """ Initialize the class-attributes in order.

        This ensures that the values are in the correct order when output.
        """
        self.name = ''
        self.enabled = True
        self.types = [util.AptSourceType.BINARY]
        self.uris = []
        self.suites = []
        self.components = []
        self.options = {}

    def make_debline(self):
        """ Output a one-line entry for this source.

        Note that this is expected to fail if somehow there is more than one
        type, URI, or suite, because this format does not support multiples of
        these items.
        """
        line = ''

        if len(self.uris) > 1:
            raise SourceError(
                'The source has too many URIs. One-line format sources support '
                'one URI only.'
            )
        if len(self.suites) > 1:
            raise SourceError(
                'The source has too many suites. One-line format sources support '
                'one suite only.'
            )
        if len(self.uris) > 1:
            raise SourceError(
                'The source has too many types. One-line format sources support '
                'one type only.'
            )

        if self.enabled == util.AptSourceEnabled.FALSE:
            line += '# '

        line += f'{self.types[0].value} '

        if self.options:
            line += '['
            line += self._get_options()
            line = line.strip()
            line += '] '

        line += f'{self.uris[0]} '
        line += f'{self.suites[0]} '

        for component in self.components:
            line += f'{component} '

        return line.strip()

    def _set_filename_ident(self, filename, ident):
        if filename:
            self.filename = filename
        if ident:
            self.ident = ident

    # pylint: disable=no-self-use
    # Handy to have for subclasses
    def _clean_name(self, name):
        name = name.replace('.list', '')
        name = name.replace('.sources', '')
        return name

    @property
    def name(self):
        """ str: The name of the source."""
        try:
            return self['X-Repolib-Name']
        except KeyError:
            return ""

    @name.setter
    def name(self, name):
        self['X-Repolib-Name'] = name

    @property
    def ident(self):
        """str: a system-scope identifier for this source. Used for filename."""
        try:
            return self._ident
        except AttributeError:
            self._ident = None
            return self._ident

    @ident.setter
    def ident(self, ident):
        self._ident = ident

    @property
    def filename(self):
        """str: The filename of the source."""
        return f'{self.ident}.sources'

    @filename.setter
    def filename(self, name):
        self._ident = self._clean_name(name)

    @property
    def enabled(self):
        """ util.AptSourceEnabled: Whether the source is enabled or not. """
        # The source can't be enabled if there are no uris or suites
        no_uris = len(self.uris) <= 0
        no_suites = len(self.suites) <= 0
        if no_uris or no_suites:
            return util.AptSourceEnabled.FALSE

        try:
            return util.AptSourceEnabled(self['enabled'])
        except KeyError:
            return ''

    @enabled.setter
    def enabled(self, enable):
        """ Accept a wide variety of data types/values for ease of use. """
        e_values = [True,
                    'True',
                    'true',
                    'Yes',
                    'yes',
                    'YES',
                    'y',
                    'Y',
                    util.AptSourceEnabled.TRUE,
                    1]
        if enable in e_values:
            self['Enabled'] = util.AptSourceEnabled.TRUE.value
        else:
            self['Enabled'] = util.AptSourceEnabled.FALSE.value

    @property
    def source_code_enabled(self):
        """bool: whether source code should be enabled or not."""
        code = False
        if self.enabled == util.AptSourceEnabled.TRUE:
            if util.AptSourceType.SOURCE in self.types:
                code = True

        self._source_code_enabled = code
        return self._source_code_enabled

    @source_code_enabled.setter
    def source_code_enabled(self, enabled):
        self._source_code_enabled = enabled
        self.types = [util.AptSourceType.BINARY]
        if enabled:
            self.types.append(util.AptSourceType.SOURCE)

    @property
    def types(self):
        """ list of util.AptSourceTypes: The types of packages provided. """
        try:
            types = []
            for dtype in self['Types'].split():
                types.append(util.AptSourceType(dtype.strip()))
            return types
        except KeyError:
            return []

    @types.setter
    def types(self, types):
        output_types = []
        for dtype in types:
            output_types.append(dtype.value)
        self['Types'] = ' '.join(output_types)


    @property
    def uris(self):
        """ [str]: The list of URIs providing packages. """
        try:
            return self['URIs'].split()
        except KeyError:
            return []

    @uris.setter
    def uris(self, uris):
        """ If the user tries to remove the last URI, disable as well. """
        if len(uris) > 0:
            for uri in uris:
                if not util.url_validator(uri):
                    raise SourceError(f'The URI "{uri}" is malformed.')
            self['URIs'] = ' '.join(uris)
        else:
            self['URIs'] = ''
            # self.enabled = False

    @property
    def suites(self):
        """ [str]: The list of enabled Suites. """
        try:
            return self['Suites'].split()
        except KeyError:
            return []

    @suites.setter
    def suites(self, suites):
        """ If user removes the last suite, disable as well. """
        if len(suites) > 0:
            self['Suites'] = ' '.join(suites)
        else:
            self['Suites'] = ''
            # self.enabled = False

    @property
    def components(self):
        """[str]: The list of components enabled. """
        try:
            return self['Components'].split()
        except KeyError:
            return []

    @components.setter
    def components(self, components):
        """ Also disable if the user tries to remove the last component. """
        if len(components) > 0:
            self['Components'] = ' '.join(components)
        else:
            self['Components'] = ''
            # self.enabled = False

    @property
    def options(self):
        """ dict: Addtional options for the repository."""
        non_options = [
            'X-Repolib-Name', 'Enabled', 'Types', 'URIs', 'Suites', 'Components'
        ]
        options = {}
        for key in self:
            if key not in non_options:
                options[key] = self[key]
        if len(options) > 0:
            return options
        return {}

    @options.setter
    def options(self, options):
        for key in options:
            self[key] = options[key]

    def _copy(self, new_source, source_code=False):
        new_source.name = self.name
        new_source.enabled = self.enabled
        new_source.types = self.types.copy()

        if source_code:
            new_source.types = [util.AptSourceType.SOURCE]

        new_source.uris = self.uris.copy()
        new_source.suites = self.suites.copy()
        new_source.components = self.components.copy()

        try:
            new_source.options = self.options.copy()
        except AttributeError:
            pass
        return new_source

    def _get_options(self):
        opt_str = ''
        for key in self.options:
            opt_str += f'{self.outoptions_d[key]}={self.options[key].replace(" ", ",")} '
        return opt_str



#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

This is a library for parsing deb lines into deb822-format data.
"""
# pylint: disable=too-many-ancestors, too-many-instance-attributes
# If we want to use the subclass, we don't have a lot of options.

from . import source
from . import util

class DebLineSourceException(Exception):
    """ Exceptions with DebLine Sources. """

    def __init__(self, *args, code=1, **kwargs):
        """Exception with a debline source object

        Arguments:
            code (:obj:`int`, optional, default=1): Exception error code.
    """
        super().__init__(*args, **kwargs)
        self.code = code

class DebLine(source.Source):
    """ Sources input via a deb line. """

    def __init__(self, line):
        super().__init__()

        self.deb_line = line
        if 'cdrom:' in self.deb_line:
            raise DebLineSourceException(
                'RepoLib does not support \'cdrom:\' URIs via this DebLine Class. '
                'Please use a Source() class to add these sources'
            )
        self._parse_debline(self.deb_line)
        self.filename = self.make_name(prefix="deb-")
        self.name = self.ident

    def copy(self, source_code=True):
        """ Copies the source and returns an identical source object.

        Arguments:
            source_code (bool): if True, output an identical source, except with
                source code enabled.

        Returns:
            A Source() object identical to self.
        """
        new_source = DebLine(self.deb_line)
        new_source = self._copy(new_source, source_code=source_code)
        return new_source

    def save_to_disk(self, save=True):
        """
        Saves the repo to disk
        """
        if save:
            super().save_to_disk(save=save)

    def _parse_debline(self, line):
        self.init_values()

        self._validate(line)
        # Enabled vs. Disabled
        self.enabled = True
        if line.startswith('#'):
            self.enabled = False
            line = line.replace('#', '', 1)
            line = line.strip()

        # URI parsing
        for word in line.split():
            if util.url_validator(word):
                self.uris = [word]
                line_uri = line.replace(word, '')

        # Options parsing
        try:
            options = self.options_re.search(line_uri).group()
            opts = self._set_options(options.strip())
            self.options = opts.copy()
            line_uri = line_uri.replace(options, '')
        except AttributeError:
            pass

        deb_list = line_uri.split()

        # Type Parsing
        self.types = [util.AptSourceType.BINARY]
        if deb_list[0] == 'deb-src':
            self.types = [util.AptSourceType.SOURCE]

        # Suite Parsing
        self.suites = [deb_list[1]]

        # Components parsing
        comps = []
        for item in deb_list[2:]:
            if not item.startswith('#'):
                comps.append(item)
            else:
                break
        self.components = comps


    def _validate(self, valid):
        """
        Ensure we have a valid debian repository line.
        """
        if valid.startswith('#'):
            valid = valid.replace('#', '')
        valid = valid.strip()
        if not valid.startswith('deb'):
            raise util.RepoError(
                'The line %s does not appear to be a valid repo' % self.deb_line
            )

    def _set_type(self, deb_type):
        """
        Set the type of repository (deb or deb-src)
        """
        self.types = [util.AptSourceType(deb_type)]

    def _set_options(self, options):
        """
        Set the options.
        """
        # Split the option string into a list of chars, so that we can replace
        # the first and last characters ([ and ]) with spaces.
        ops = list(options)
        ops[0] = " "
        ops[-1] = " "
        options = "".join(ops).strip()

        for replacement in self.options_d:
            options = options.replace(replacement, self.options_d[replacement])

        options = options.replace('=', ',')
        options_list = options.split()

        options_output = {}

        for i in options_list:
            option = i.split(',')
            values_list = []
            for value in option[1:]:
                values_list.append(value)
            options_output[option[0]] = ' '.join(values_list)
        return options_output



#!/usr/bin/python3
'''
   Copyright 2020 Ian Santopietro (ian@system76.com)

   This file is part of Repolib.

    Repolib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Repolib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Repolib.  If not, see <http://www.gnu.org/licenses/>.
'''
#pylint: skip-file

import gi
from gi.repository import GObject, GLib

from pathlib import Path
import dbus
import dbus.service
import dbus.mainloop.glib
import time

import repolib

class RepolibException(dbus.DBusException):
    _dbus_error_name = 'org.pop_os.repolib.RepolibException'

class PermissionDeniedByPolicy(dbus.DBusException):
    _dbus_error_name = 'org.pop_os.repolib.PermissionDeniedByPolicy'

class AptException(Exception):
    pass

class Repo(dbus.service.Object):
    def __init__(self, conn=None, object_path=None, bus_name=None):
        dbus.service.Object.__init__(self, conn, object_path, bus_name)

        # These are used by PolKit to check privileges
        self.dbus_info = None
        self.polkit = None
        self.enforce_polkit = True

        try:
            self.system_repo = repolib.SystemSource()
        except:
            self.system_repo = None
        
        self.source = None
        self.sources_dir = Path('/etc/apt/sources.list.d')
    
    @dbus.service.method(
        "org.pop_os.repolib.Interface",
        in_signature='ss', out_signature='',
        sender_keyword='sender', connection_keyword='conn'
    )
    def output_file_to_disk(self, filename, source, sender=None, conn=None):
        self._check_polkit_privilege(
            sender, conn, 'org.pop_os.repolib.modifysources'
        )
        full_path = self.sources_dir / filename
        with open(full_path, mode='w') as output_file:
            output_file.write(source)

    @dbus.service.method(
        "org.pop_os.repolib.Interface",
        in_signature='s', out_signature='',
        sender_keyword='sender', connection_keyword='conn'
    )
    def delete_source(self, filename, sender=None, conn=None):
        self._check_polkit_privilege(
            sender, conn, 'org.pop_os.repolib.modifysources'
        )
        source_file = self.sources_dir / filename
        source_file.unlink()

    @dbus.service.method(
        "org.pop_os.repolib.Interface",
        in_signature='', out_signature='',
        sender_keyword='sender', connection_keyword='conn'
    )
    def exit(self, sender=None, conn=None):
        mainloop.quit()
    
    @dbus.service.method(
        "org.pop_os.repolib.Interface",
        in_signature='b', out_signature='b',
        sender_keyword='sender', connection_keyword='conn'
    )
    def set_system_source_code_enabled(self, enabled, sender=None, conn=None):
        """ Enable or disable source code in the system source. 
        
        Arguments:
            enabled (bool): The new state to set, True = Enabled.
        """
        self._check_polkit_privilege(
            sender, conn, 'org.pop_os.repolib.modifysources'
        )
        if self.system_repo:
            self.system_repo.load_from_file()
            new_types = [repolib.util.AptSourceType.BINARY]
            if enabled:
                new_types.append(repolib.util.AptSourceType.SOURCE)
            self.system_repo.types = new_types
            self.system_repo.save_to_disk()
            return enabled
        return False

    @dbus.service.method(
        "org.pop_os.repolib.Interface",
        in_signature='sb', out_signature='b',
        sender_keyword='sender', connection_keyword='conn'
    )
    def set_system_comp_enabled(self, comp, enable, sender=None, conn=None):
        """ Enable or disable a component in the system source. 
        
        Arguments:
            comp (str): the component to set
            enable (bool): The new state to set, True = Enabled.
        """
        self._check_polkit_privilege(
            sender, conn, 'org.pop_os.repolib.modifysources'
        )
        if self.system_repo:
            self.system_repo.load_from_file()
            self.system_repo.set_component_enabled(component=comp, enabled=enable)
            self.system_repo.save_to_disk()
            return True
        return False
    
    @dbus.service.method(
        "org.pop_os.repolib.Interface",
        in_signature='sb', out_signature='b',
        sender_keyword='sender', connection_keyword='conn'
    )
    def set_system_suite_enabled(self, suite, enable, sender=None, conn=None):
        """ Enable or disable a suite in the system source. 
        
        Arguments:
            suite (str): the suite to set
            enable (bool): The new state to set, True = Enabled.
        """
        self._check_polkit_privilege(
            sender, conn, 'org.pop_os.repolib.modifysources'
        )
        if self.system_repo:
            self.system_repo.load_from_file()
            self.system_repo.set_suite_enabled(suite=suite, enabled=enable)
            self.system_repo.save_to_disk()
            return True
        return False

    @classmethod
    def _log_in_file(klass, filename, string):
        date = time.asctime(time.localtime())
        ff = open(filename, "a")
        ff.write("%s : %s\n" %(date,str(string)))
        ff.close()
    
    @classmethod
    def _strip_source_line(self, source):
        source = source.replace("#", "# ")
        source = source.replace("[", "")
        source = source.replace("]", "")
        source = source.replace("'", "")
        source = source.replace("  ", " ")
        return source

    def _check_polkit_privilege(self, sender, conn, privilege):
        # from jockey
        '''Verify that sender has a given PolicyKit privilege.
        sender is the sender's (private) D-BUS name, such as ":1:42"
        (sender_keyword in @dbus.service.methods). conn is
        the dbus.Connection object (connection_keyword in
        @dbus.service.methods). privilege is the PolicyKit privilege string.
        This method returns if the caller is privileged, and otherwise throws a
        PermissionDeniedByPolicy exception.
        '''
        if sender is None and conn is None:
            # called locally, not through D-BUS
            return
        if not self.enforce_polkit:
            # that happens for testing purposes when running on the session
            # bus, and it does not make sense to restrict operations here
            return

        # get peer PID
        if self.dbus_info is None:
            self.dbus_info = dbus.Interface(conn.get_object('org.freedesktop.DBus',
                '/org/freedesktop/DBus/Bus', False), 'org.freedesktop.DBus')
        pid = self.dbus_info.GetConnectionUnixProcessID(sender)
        
        # query PolicyKit
        if self.polkit is None:
            self.polkit = dbus.Interface(dbus.SystemBus().get_object(
                'org.freedesktop.PolicyKit1',
                '/org/freedesktop/PolicyKit1/Authority', False),
                'org.freedesktop.PolicyKit1.Authority')
        try:
            # we don't need is_challenge return here, since we call with AllowUserInteraction
            (is_auth, _, details) = self.polkit.CheckAuthorization(
                    ('unix-process', {'pid': dbus.UInt32(pid, variant_level=1),
                    'start-time': dbus.UInt64(0, variant_level=1)}), 
                    privilege, {'': ''}, dbus.UInt32(1), '', timeout=600)
        except dbus.DBusException as e:
            if e._dbus_error_name == 'org.freedesktop.DBus.Error.ServiceUnknown':
                # polkitd timed out, connect again
                self.polkit = None
                return self._check_polkit_privilege(sender, conn, privilege)
            else:
                raise

        if not is_auth:
            Repo._log_in_file('/tmp/repolib.log','_check_polkit_privilege: sender %s on connection %s pid %i is not authorized for %s: %s' %
                    (sender, conn, pid, privilege, str(details)))
            raise PermissionDeniedByPolicy(privilege)

if __name__ == '__main__':
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    
    bus = dbus.SystemBus()
    name = dbus.service.BusName("org.pop_os.repolib", bus)
    object = Repo(bus, '/Repo')

    mainloop = GLib.MainLoop()
    mainloop.run()


#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

This is a library for parsing deb lines into deb822-format data.
"""
# pylint: disable=missing-function-docstring, missing-class-docstring
# These aren't really relevant for unit testing (which is mostly automatic.)

import unittest

from . import deb
from . import util

class DebTestCase(unittest.TestCase):
    def test_normal_source(self):
        source = deb.DebLine(
            'deb http://example.com/ suite main'
        )
        self.assertEqual(source.name, 'deb-example-com')
        self.assertEqual(source.types, [util.AptSourceType.BINARY])
        # pylint: disable=no-member
        # This works fine, so I think pylint is just confused.
        self.assertTrue(source.enabled.get_bool())
        # pylint: enable=no-member
        self.assertEqual(source.uris, ['http://example.com/'])
        self.assertEqual(source.suites, ['suite'])
        self.assertEqual(source.components, ['main'])
        self.assertFalse(source.options)
        self.assertEqual(source.ident, 'deb-example-com')

    def test_source_with_multiple_components(self):
        source = deb.DebLine(
            'deb http://example.com/ suite main nonfree'
        )
        self.assertEqual(source.suites, ['suite'])
        self.assertEqual(source.components, ['main', 'nonfree'])

    def test_source_with_option(self):
        source = deb.DebLine(
            'deb [ arch=amd64 ] http://example.com/ suite main'
        )
        self.assertEqual(source.uris, ['http://example.com/'])
        self.assertDictEqual(source.options, {'Architectures': 'amd64'})

    def test_source_uri_with_brackets(self):
        source = deb.DebLine(
            'deb http://example.com/[release]/ubuntu suite main'
        )
        self.assertEqual(source.uris, ['http://example.com/[release]/ubuntu'])
        self.assertFalse(source.options, {})

    def test_source_options_with_colons(self):
        source = deb.DebLine(
            'deb [ arch=arm:2 ] http://example.com/ suite main'
        )
        self.assertEqual(source.uris, ['http://example.com/'])
        self.assertDictEqual(source.options, {'Architectures': 'arm:2'})

    def test_source_with_multiple_option_values(self):
        source = deb.DebLine(
            'deb [ arch=armel,amd64 ] http://example.com/ suite main'
        )
        self.assertEqual(source.uris, ['http://example.com/'])
        self.assertDictEqual(source.options, {'Architectures': 'armel amd64'})

    def test_source_with_multiple_options(self):
        source = deb.DebLine(
            'deb [ arch=amd64 lang=en_US ] http://example.com/ suite main'
        )
        self.assertEqual(source.uris, ['http://example.com/'])
        self.assertDictEqual(
            source.options,
            {'Architectures': 'amd64', 'Languages': 'en_US'}
        )

    def test_source_with_multiple_options_with_multiple_values(self):
        source = deb.DebLine(
            'deb [ arch=amd64,armel lang=en_US,en_CA ] '
            'http://example.com/ suite main'
        )
        self.assertEqual(source.uris, ['http://example.com/'])
        self.assertDictEqual(
            source.options,
            {'Architectures': 'amd64 armel', 'Languages': 'en_US en_CA'}
        )

    def test_source_uri_with_brackets_and_options(self):
        source = deb.DebLine(
            'deb [ arch=amd64 lang=en_US,en_CA ] '
            'http://example][.com/[release]/ubuntu suite main'
        )
        self.assertEqual(source.uris, ['http://example][.com/[release]/ubuntu'])
        self.assertDictEqual(
            source.options,
            {'Architectures': 'amd64', 'Languages': 'en_US en_CA'}
        )

    def test_source_uri_with_brackets_and_options_with_colons(self):
        source = deb.DebLine(
            'deb [ arch=amd64,arm:2 lang=en_US,en_CA ] '
            'http://example][.com/[release]/ubuntu suite main'
        )
        self.assertEqual(source.uris, ['http://example][.com/[release]/ubuntu'])
        self.assertDictEqual(
            source.options,
            {'Architectures': 'amd64 arm:2', 'Languages': 'en_US en_CA'}
        )

    def test_worst_case_sourcenario(self):
        source = deb.DebLine(
            'deb [ arch=amd64,arm:2,arm][ lang=en_US,en_CA ] '
            'http://example][.com/[release:good]/ubuntu suite main restricted '
            'nonfree not-a-component'
        )
        self.assertEqual(source.uris, ['http://example][.com/[release:good]/ubuntu'])
        self.assertEqual(source.suites, ['suite'])
        self.assertEqual(source.components, [
            'main', 'restricted', 'nonfree', 'not-a-component'
        ])
        self.assertDictEqual(
            source.options,
            {
                'Architectures': 'amd64 arm:2 arm][',
                'Languages': 'en_US en_CA'
            }
        )

    def test_source_code_source(self):
        source = deb.DebLine(
            'deb-src http://example.com/ suite main'
        )
        self.assertEqual(source.name, 'deb-example-com')
        self.assertEqual(source.types, [util.AptSourceType.SOURCE])

    def test_disabled_source(self):
        source = deb.DebLine(
            '# deb http://example.com/ suite main'
        )
        self.assertEqual(source.name, 'deb-example-com')
        # pylint: disable=no-member
        # This works fine, so I think pylint is just confused.
        self.assertFalse(source.enabled.get_bool())
        # pylint: enable=no-member

    def test_disabled_source_without_space(self):
        source = deb.DebLine(
            '#deb http://example.com/ suite main'
        )
        self.assertEqual(source.name, 'deb-example-com')
        # pylint: disable=no-member
        # This works fine, so I think pylint is just confused.
        self.assertFalse(source.enabled.get_bool())
        # pylint: enable=no-member

    def test_source_with_trailing_comment(self):
        source = deb.DebLine(
            'deb http://example.com/ suite main # This is a comment'
        )
        self.assertEqual(source.name, 'deb-example-com')
        self.assertEqual(source.suites, ['suite'])
        self.assertEqual(source.components, ['main'])

    def test_disabled_source_with_trailing_comment(self):
        source = deb.DebLine(
            '# deb http://example.com/ suite main # This is a comment'
        )
        self.assertEqual(source.name, 'deb-example-com')
        self.assertEqual(source.suites, ['suite'])
        # pylint: disable=no-member
        # This works fine, so I think pylint is just confused.
        self.assertFalse(source.enabled.get_bool())
        # pylint: enable=no-member
        self.assertEqual(source.components, ['main'])

    @unittest.expectedFailure
    def test_cdrom_source(self):
        source = deb.DebLine(
            '# deb cdrom:[This is a CD-ROM Source] suite main # This is a comment'
        )
        self.assertEqual(source.uris, ['cdrom:[This is a CD-ROM Source]'])




"""
Copyright (c) 2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

Command to remove sources from the system.
"""

from pathlib import Path

import dbus

from . import command
from ..util import get_sources_dir, RepoError

class Remove(command.Command):
    # pylint: disable=no-self-use,too-few-public-methods
    # This is a base class for other things to inherit and give other programs
    # a standardized interface for interacting with commands.
    """ Remove subcommand.

    The remove command will remove the selected source. It has no options. Note
    that the system sources cannot be removed. This requires root.
    """

    @classmethod
    def init_options(cls, subparsers):
        """ Sets up this command's options parser.

        Returns:
            The subparser for this command.
        """
        parser_remove = subparsers.add_parser(
            'remove',
            help='Remove a configured repository.'
        )
        parser_remove.add_argument(
            'repository',
            help='The name of the repository to remove. See LIST'
        )

    def __init__(self, log, args, parser):
        super().__init__(log, args, parser)
        self.source = args.repository
        self.sources_dir = get_sources_dir()

    def get_source_path(self):
        """ Tries to get the full path to the source.

        This is necessary because some sources end in .list, others in .sources

        Returns:
            A tuple with the pathlib.Path to the sources, and the pathlib.Path
            to the '.save' file.
        """
        full_name = f'{self.source}.sources'
        full_path = self.sources_dir / full_name
        self.log.debug('Trying to load %s', full_path)
        if full_path.exists():
            return full_path, full_path

        full_name = f'{self.source}.list'
        full_path = self.sources_dir / full_name
        self.log.debug('Trying to load %s', full_path)
        if full_path.exists():
            save_path = Path(f'{full_path}.save')
            return full_path, save_path

        raise RepoError('The path does not exist (checked .sources and .list files.')

    def run(self):
        """ Run the command. """

        if self.source.lower() == 'system':
            self.log.error("You cannot remove the system sources!")
            return False

        try:
            remove_path, remove_path_save = self.get_source_path()
        except RepoError:
            self.log.error(
                'No source %s found on system. Check the spelling.', self.source
            )
            return False

        print(
            f'You are about to remove the sources contained in {remove_path.name}.'
            '\nAre you sure you want to do this?\n'
        )
        response = 'f'
        while response.lower() not in ['y', 'n']:
            response = input(f'Remove {remove_path.name}? (y/N) ')
            if response == '':
                response = 'n'

        if response.lower() == 'y':
            if self.args.debug != 0:
                self.log.info('Simulate: Remove %s', remove_path)
                self.log.info('Simulate: Remove %s', remove_path_save)
            else:
                try:
                    remove_path.unlink()
                    remove_path_save.unlink(missing_ok=True)
                except PermissionError:
                    bus = dbus.SystemBus()
                    privileged_object = bus.get_object('org.pop_os.repolib', '/Repo')
                    privileged_object.delete_source(remove_path.name)
                    privileged_object.exit()

        else:
            print('Canceled.')
            return False

        return True



#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=too-many-ancestors, too-many-instance-attributes
# If we want to use the subclass, we don't have a lot of options.

import dbus

from . import deb
from . import source
from . import util

def combine_lists(list1, list2):
    """ Adds list2 to list1, without adding duplicates.

    Arguments:
        list1 (list): The list to add to
        list2 (list): The list to add from

    Returns:
        A list of the two combined.
    """
    for item in list2:
        if item not in list1:
            list1.append(item)

    return list1

class LegacyDebSource(source.Source):
    """Legacy deb sources

    Because Legacy Deb format entries have limitations on how many URIs or
    suites they can contain, many legacy entry files use multiple sources to
    configure multiple URIs, suites, or types. A common example is to have a
    `deb` entry and an otherwise identical `deb-src` entry. To make this
    simpler, we treat legacy sources as a "meta source" and store the individual
    lines in a list.

    Keyword Arguments:
        name (str): The name of this source
        filename (str): The name of the source file on disk
    """

    # pylint: disable=super-init-not-called
    # Because this is a sort of meta-source, it needs to be different from the
    # super class.
    def __init__(self, *args, filename=None, ident=None, **kwargs):
        super().__init__(*args, filename=filename, ident=ident, **kwargs)
        self.sources = []
        self._source_code_enabled = False
        self.init_values()

    def make_names(self):
        """ Creates a filename for this source, if one is not provided.

        It also sets these values up.
        """

        if len(self.sources) <= 0:
            self.name = self.make_name()

        if not self.ident:
            self.ident = self.sources[0].make_name()

        if not self.name:
            self.name = self.sources[0].make_name()
            self.name = self.name.replace('.sources', '')

    def load_from_sources(self):
        """ Loads the source from its consituent source lines."""
        enabled = False
        uris = []
        suites = []
        components = []
        options = {}

        for repo in self.sources:
            if repo.types[0] not in self.types:
                self.types.append(repo.types[0])
            if repo.enabled.value == 'yes':
                if util.AptSourceType.BINARY in repo.types:
                    enabled = True
                else:
                    self.source_code_enabled = True
            uris = combine_lists(uris, repo.uris)
            suites = combine_lists(suites, repo.suites)
            components = combine_lists(components, repo.components)
            options.update(repo.options)

        self.uris = uris.copy()
        self.suites = suites.copy()
        self.components = components.copy()
        self.options = options.copy()
        self.enabled = enabled
        self.enabled = self.enabled

        if not self.name:
            self.make_names()
            for src in self.sources:
                src.name = self.name

    def load_from_file(self, filename=None, ident=None):
        """ Loads the source from a file on disk.

        Keyword arguments:
          filename -- STR, Containing the path to the file. (default: self.filename)
        """
        self._set_filename_ident(filename, ident)
        self.sources = []
        name = None

        full_path = util.get_sources_dir() / self.filename
        sources = []
        with open(full_path, 'r') as source_file:
            for line in source_file:
                if util.validate_debline(line):
                    deb_src = deb.DebLine(line)
                    deb_src.name = self.name
                    sources.append(deb_src)
                elif "X-Repolib-Name" in line:
                    name = ':'.join(line.split(':')[1:])
                    self.name = name.strip()

        self.sources = sources
        self.load_from_sources()

    # pylint: disable=arguments-differ
    # This is operating on a very different kind of source, thus needs to be
    # different.
    def save_to_disk(self):
        """ Save the source to the disk. """
        try:
            self.sources[0].save_to_disk(save=False)
        except IndexError:
            pass
        full_path = util.get_sources_dir() / self.filename

        source_output = self.make_deblines()
        try:
            with open(full_path, 'w') as source_file:
                source_file.write(source_output)
        except PermissionError:
            bus = dbus.SystemBus()
            privileged_object = bus.get_object('org.pop_os.repolib', '/Repo')
            privileged_object.output_file_to_disk(self.filename, source_output)
            privileged_object.exit()

    def make_deblines(self):
        """ Create a string representation of the enties as they would be saved.

        This is useful for testing, and is used by the save_to_disk() method.

        Returns:
            A str with the output entries.
        """
        toprint = '## Added/managed by repolib ##\n'
        toprint += f'#\n## X-Repolib-Name: {self.name}\n'

        for suite in self.suites:
            for uri in self.uris:
                out_binary = source.Source()
                out_binary.name = self.name
                out_binary.enabled = self.enabled.value
                out_binary.types = [util.AptSourceType.BINARY]
                out_binary.uris = [uri]
                out_binary.suites = [suite]
                out_binary.components = self.components
                out_binary.options = self.options
                toprint += f'{out_binary.make_debline()}\n'

                out_source = out_binary.copy()
                out_source.enabled = self.source_code_enabled
                out_source.types = [util.AptSourceType.SOURCE]
                toprint += f'{out_source.make_debline()}\n'

        return toprint


    @property
    def source_code_enabled(self):
        """bool: whether source code should be enabled or not."""
        source_code = False
        for repo in self.sources:
            if repo.enabled == util.AptSourceEnabled.TRUE:
                if util.AptSourceType.SOURCE in repo.types:
                    source_code = True
        self._source_code_enabled = source_code
        return self._source_code_enabled

    @source_code_enabled.setter
    def source_code_enabled(self, enabled):
        """This needs to be tracked somewhat separately"""
        self._source_code_enabled = enabled
        for repo in self.sources:
            if util.AptSourceType.SOURCE in repo.types and self.enabled:
                repo.enabled = enabled

    @property
    def filename(self):
        """str: The filename of the source."""
        return f'{self.ident}.list'

    @filename.setter
    def filename(self, name):
        self._ident = self._clean_name(name)

    @property
    def types(self):
        """ list of util.AptSourceTypes: The types of packages provided.

        We need to override in order to learn this from the source_code_enabled
        property.
        """
        if self.source_code_enabled:
            self['Types'] = 'deb deb-src'
        else:
            self['Types'] = 'deb'

        types = []
        try:
            for dtype in self['Types'].split():
                types.append(util.AptSourceType(dtype.strip()))
            return types

        except KeyError:
            return []

    @types.setter
    def types(self, types):
        if util.AptSourceType.SOURCE in types:
            self.source_code_enabled = True
        else:
            self.source_code_enabled = False

        output_types = []
        for dtype in types:
            output_types.append(dtype.value)
        self['Types'] = ' '.join(output_types)



#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

This is a library for parsing deb lines into deb822-format data.
"""
# pylint: disable=missing-function-docstring, missing-class-docstring
# These aren't really relevant for unit testing (which is mostly automatic.)

import unittest

from . import source
from . import util

class SourceTestCase(unittest.TestCase):
    def setUp(self):
        self.source = source.Source(ident='test')
        self.source.name = 'Test Source'
        self.source.enabled = False
        self.source.types = [util.AptSourceType.BINARY, util.AptSourceType.SOURCE]
        self.source.uris = ['http://example.com/ubuntu', 'http://example.com/mirror']
        self.source.suites = ['suite', 'suite-updates']
        self.source.components = ['main', 'contrib', 'nonfree']
        self.source.options = {
            'Architectures': 'amd64 armel',
            'Languages': 'en_US en_CA'
        }

    def test_default_source_data(self):
        self.assertEqual(self.source.name, 'Test Source')
        # pylint: disable=no-member
        # This works fine, so I think pylint is just confused.
        self.assertFalse(self.source.enabled.get_bool())
        self.assertEqual(self.source.types, [
            util.AptSourceType.BINARY, util.AptSourceType.SOURCE
        ])
        self.assertEqual(self.source.uris, [
            'http://example.com/ubuntu', 'http://example.com/mirror'
        ])
        self.assertEqual(self.source.suites, [
            'suite', 'suite-updates'
        ])
        self.assertEqual(self.source.components, [
            'main', 'contrib', 'nonfree'
        ])
        self.assertDictEqual(self.source.options, {
            'Architectures': 'amd64 armel',
            'Languages': 'en_US en_CA'
        })
        self.assertEqual(self.source.filename, 'test.sources')

    def test_make_source_string(self):
        source_string = (
            'Name: Test Source\n'
            'Enabled: no\n'
            'Types: deb deb-src\n'
            'URIs: http://example.com/ubuntu http://example.com/mirror\n'
            'Suites: suite suite-updates\n'
            'Components: main contrib nonfree\n'
            'Architectures: amd64 armel\n'
            'Languages: en_US en_CA\n'
        )
        self.assertEqual(self.source.make_source_string(), source_string)

    def test_set_enabled(self):
        self.source.enabled = True
        # pylint: disable=no-member
        # This works fine, so I think pylint is just confused.
        self.assertTrue(self.source.enabled.get_bool())

    def test_set_source_enabled(self):
        self.source.set_source_enabled(False)
        self.assertEqual(self.source.types, [util.AptSourceType.BINARY])



#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.
"""
#pylint: disable=too-many-ancestors, too-many-instance-attributes
# If we want to use the subclass, we don't have a lot of options.

from . import source
from . import util

class SystemSourceException(Exception):
    """ System Source Exceptions. """

    def __init__(self, *args, code=1, **kwargs):
        """Exception with the system sources

        Arguments:
            msg (str): Human-readable message describing the error that threw the
                exception.
            code (:obj:`int`, optional, default=1): Exception error code.
    """
        super().__init__(*args, **kwargs)
        self.code = code

class SystemSource(source.Source):
    """ System Sources. """

    def __init__(self, ident='system'):
        """ Constructor for System Sources

        Loads a source object for the System Sources. These are located (by
        default) in /etc/apt/sources.list.d/system.sources. If your distro uses
        a different location, please patch this in your packaging.
        """
        super().__init__()
        self.ident = ident
        self.load_from_file()

    def set_component_enabled(self, component='main', enabled=True):
        """ Enables or disabled a repo component (e.g. 'main')

        Keyword Arguments:
            component -- The component to (en|dis)able (default: "main")
            ennabled -- Whether COMPONENT is enabled (default: True)
        """
        components = self.components.copy()
        if not enabled:
            if component in components:
                components.remove(component)
                self.components = components.copy()
                self.save_to_disk()
                return component
        else:
            if component not in components:
                self.enabled = True
                components.append(component)
                self.components = components.copy()
                self.save_to_disk()
                return component

        raise SystemSourceException(
            msg=f"Couldn't toggle component: {component} to {enabled}"
        )

    def set_suite_enabled(self, suite=util.DISTRO_CODENAME, enabled=True):

        """ Enables or disabled a repo suite (e.g. 'main')

        Keyword Arguments:
            suite -- The suite to (en|dis)able (default: main)
            ennabled -- Whether COMPONENT is enabled (default: True)
        """
        suites = self.suites.copy()
        if not enabled:
            if suite in suites:
                suites.remove(suite)
                self.suites = suites.copy()
                self.save_to_disk()
                return suite
        else:
            if suite not in suites:
                self.enabled = True
                suites.append(suite)
                self.suites = suites.copy()
                self.save_to_disk()
                return suite


        raise SystemSourceException(
            msg=f"Couldn't toggle suite: {suite} to {enabled}"
        )

    def set_default_mirror(self):
        """ Resets the System Sources to use the default mirrors.

        Requires that the `default_mirror` attribute be set.
        """
        if self.default_mirror:
            self.uris = [self.default_mirror]
            return
        raise SystemSourceException('No default mirror set.')

    @property
    def default_mirror(self):
        """str: The default system mirror."""
        try:
            return self['X-Repolib-Default-Mirror']
        except KeyError:
            return ''

    @default_mirror.setter
    def default_mirror(self, uri):
        if util.url_validator(uri):
            self['X-Repolib-Default-Mirror'] = uri




"""
Copyright (c) 2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

Source Subcommand.
"""

from ..legacy_deb import LegacyDebSource
from ..source import Source as Source_obj
from ..util import get_source_path

from . import command

class Source(command.Command):
    """ Source Subcommand.

    The source command allows enabling or disabling source code in configured
    sources. If a configured source is provided, this command will affect that
    source. If no sources are provided, this command will affect all sources on
    the system. Without options, it will list the status for source
    code packages.

    Options:
        --enable, -e
        --disable, -d
    """

    @classmethod
    def init_options(cls, subparsers):
        """ Sets up this command's options parser.

        Returns:
            The subparser for this command.
        """
        parser_source = subparsers.add_parser(
            'source',
            help='Enable/disable source code packages for repositories'
        )
        parser_source.add_argument(
            'repository',
            nargs='*',
            default=['x-repolib-all-sources'],
            help=(
                'The repository for which to modify source code packages. If not '
                'specified, then attempt to modify for all repositories.'
            )
        )

        source_enable = parser_source.add_mutually_exclusive_group(
            required=True
        )
        source_enable.add_argument(
            '-e',
            '--enable',
            action='store_true',
            dest='source_enable',
            help='Enable source code for the repository'
        )
        source_enable.add_argument(
            '-d',
            '--disable',
            action='store_true',
            dest='source_disable',
            help='Disable source code for the repository'
        )

    def finalize_options(self, args):
        """ Set up options:

        args.repository [str]: The repository to modify
        args.enable | args.disable: whether to enable or disable source code.
        """
        super().finalize_options(args)
        self.repo = ' '.join(args.repository)
        self.enable = True
        if args.source_disable:
            self.enable = False

    def run(self):
        """ Run the command."""
        if self.repo == 'x-repolib-all-sources':
            self.log.error('You must specify a Repository')
            return False

        self.log.debug('Setting for repo %s: Source Code: %s', self.repo, self.enable)
        file_path = get_source_path(self.repo, log=self.log)
        if file_path.suffix == '.sources':
            repo = Source_obj(filename=file_path.name)
        elif file_path.suffix == '.list':
            repo = LegacyDebSource(filename=file_path.name)

        repo.load_from_file()

        try:
            set_source_enabled(repo, self.enable)
            if self.verbose:
                print(repo.make_source_string())
            if not self.debug:
                repo.save_to_disk()

        # pylint: disable=invalid-name
        # This is a pretty widely used convention
        except command.RepolibCommandError as e:
            self.log.error(
                'Could not change source code for %s:\n%s',
                self.repo,
                e
            )

        return True

def set_source_enabled(repo, enabled):
    """ Enables or disables Source Code for the given repository.

    Arguments:
        repo - Repolib.Source: The repository to enable or disable source for.
        enabled - bool: The state to which source code should be set.
    """
    try:
        if enabled:
            repo.types = [command.AptSourceType.BINARY, command.AptSourceType.SOURCE]
        else:
            repo.types = [command.AptSourceType.BINARY]
    # pylint: disable=invalid-name
    # This is a pretty widely used convention
    except Exception as e:
        raise command.RepolibCommandError(
            f'Could not set the source code for {repo.name} to {enabled}'
        ) from e



#!/usr/bin/python3

"""
Copyright (c) 2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

This subpackage contains tools for use in developing CLI applications using
repolib. apt-manage uses this library.
"""
#pylint: disable=invalid-name,cyclic-import
# Can't seem to find the cyclic import being complained about.

from .argparser import get_argparser
from .add import Add
from .list import List
from .modify import Modify
from .remove import Remove
from .source import Source

parser = get_argparser()



#!/usr/bin/python3

"""
Copyright (c) 2019-2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.
"""
#pylint: disable=invalid-name,subprocess-run-check
# We don't need to check these in setup

import os
import subprocess
from setuptools import setup, find_packages, Command

def get_version():
    """ Get the program version. """
    #pylint: disable=exec-used
    # Just getting the version.
    version = {}
    with open(os.path.join('repolib', '__version__.py')) as fp:
        exec(fp.read(), version)
    return version['__version__']

with open("README.rst", "r") as fh:
    long_description = fh.read()

classifiers = [
    'Environment :: Console',
    'Intended Audience :: System Administrators',
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Natural Language :: English',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python :: 3',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: System'
]

class Release(Command):
    """ Generate a release and push it to git."""
    description = "Generate a release and push it to git."

    user_options = [
        ('dry-run', None, 'Skip the actual release and do a dry run instead.'),
        ('skip-deb', None, 'Skip doing a debian update for this release.'),
        ('skip-git', None, 'Skip committing to git at the end.'),
        ('prerelease=', None, 'Release a pre-release version (alpha,beta,rc)'),
        ('increment=', None, 'Manually specify the desired increment (MAJOR, MINOR, PATCH)')
    ]

    def initialize_options(self):
        self.dry_run = False
        self.skip_deb = False
        self.skip_git = False
        self.prerelease = None
        self.increment = None

    def finalize_options(self):
        pass

    def run(self):
        cz_command = ['cz', 'bump', '--yes']
        ch_command = ['dch']
        git_command = ['git', 'add', '.']

        def capture_version(sp_complete):
            output = sp_complete.stdout.decode('UTF-8').split('\n')
            print('\n'.join(output))
            for line in output:
                if 'tag to create' in line:
                    version_line = line

            try:
                return version_line.split()[-1].replace('v', '')
            except UnboundLocalError:
                stderr = sp_complete.stderr.decode('UTF-8')
                print("WARNING: Couldn't get updated version! Using current.")
                print(stderr)
                return get_version()

        if self.dry_run:
            print('Dry run: Not making actual changes')
            cz_command.append('--dry-run')

        if self.prerelease:
            if self.prerelease.lower() not in ['alpha', 'beta', 'rc']:
                raise Exception(
                    f'{self.prerelease} is not a valid prerelease type. Please '
                    'use one of "alpha", "beta", or "rc".'
                )
            cz_command.append('--prerelease')
            cz_command.append(self.prerelease.lower())

        if self.increment:
            if self.increment.upper() not in ['MAJOR', 'MINOR', 'PATCH']:
                raise Exception(
                    f'{self.increment} is not a valid increments. Please use '
                    'one of MAJOR, MINOR, or PATCH.'
                )
            cz_command.append('--increment')
            cz_command.append(self.increment.upper())

        # We need to get the new version from CZ, as the file hasn't been
        # updated yet.
        version_command = cz_command.copy()
        version_command.append('--dry-run')
        version_complete = subprocess.run(version_command, capture_output=True)

        version = capture_version(version_complete)
        print(f'Old Version: {get_version()}')
        print(f'New version: {version}')

        ch_command.append(f'-v{version}')
        if not self.skip_deb:
            print(ch_command)
            if not self.dry_run:
                subprocess.run(ch_command)
                subprocess.run(['dch', '-r', '""'])

        if not self.skip_git:
            print(git_command)
            if not self.dry_run:
                subprocess.run(git_command)

        print(' '.join(cz_command))
        if not self.dry_run:
            subprocess.run(cz_command)

class Test(Command):
    """ Run pyflakes and pytest"""
    description = 'Run pyflakes, pytest, and pylint'

    user_options = [
        ('run-flakes', None, 'Run pyflakes'),
        ('skip-test', None, 'Skip running pytest'),
        ('skip-lint', None, 'Skip running pylint')
    ]

    def initialize_options(self):
        self.run_flakes = True
        self.skip_test = False
        self.skip_lint = False

    def finalize_options(self):
        pass

    def run(self):
        pytest_command = ['pytest-3']
        flakes_command = ['pyflakes3', 'repolib']
        lint_command = ['pylint', 'repolib']

        if not self.skip_test:
            subprocess.run(pytest_command)

        if not self.run_flakes:
            subprocess.run(flakes_command)

        if not self.skip_lint:
            subprocess.run(lint_command)

setup(
    name='repolib',
    version=get_version(),
    author='Ian Santopietro',
    author_email='ian@system76.com',
    url='https://github.com/pop-os/repolib',
    description='Easily manage software sources',
    download_url='https://github.com/pop-os/repolib/releases',
    long_description=long_description,
    tests_require=['pytest'],
    license='LGPLv3',
    packages=['repolib', 'repolib/command'],
    cmdclass={'release': Release, 'test': Test},
    scripts=['bin/apt-manage'],
    data_files=[
        ('share/bash-completion/completions', ['data/bash-completion/apt-manage']),
        ('share/zsh/vendor-completions', ['data/zsh-completion/_apt-manage']),
        ('/usr/share/dbus-1/system-services', ['data/org.pop_os.repolib.service']),
        ('/usr/share/polkit-1/actions', ['data/org.pop_os.repolib.policy']),
        ('/etc/dbus-1/system.d/', ['data/org.pop_os.repolib.conf']),
        ('/usr/lib/repolib', ['data/service.py', 'bin/add-apt-repository'])
    ]
)



# pylint: disable=unused-import,cyclic-import
# It isn't used by the base class, but is used by subclasses.
# Can't seem to find the cyclic import being complained about.

"""
Copyright (c) 2020, Ian Santopietro
All rights reserved.

This file is part of RepoLib.

RepoLib is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RepoLib is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with RepoLib.  If not, see <https://www.gnu.org/licenses/>.

Command line application helper class.
"""

from ..util import get_sources_dir, AptSourceType

class RepolibCommandError(Exception):
    """ Exceptions generated by Repolib CLIs."""

    def __init__(self, *args, code=1, **kwargs):
        """Exception from a CLI

        Arguments:
            code (:obj:`int`, optional, default=1): Exception error code.
    """
        super().__init__(*args, **kwargs)
        self.code = code

class Command:
    # pylint: disable=no-self-use,too-few-public-methods
    # This is a base class for other things to inherit and give other programs
    # a standardized interface for interacting with commands.
    """ CLI helper class for developing command line applications."""

    def __init__(self, log, args, parser):
        self.log = log
        self.parser = parser
        self.sources_dir = get_sources_dir()
        self.finalize_options(args)

    def finalize_options(self, args):
        """ Base options parsing class.

        Use this class in commands to set up the final options parsing and set
        instance variables for later use.
        """
        # This is set for backwards-compatibility with existing commands. It
        # will be removed in future versions.
        self.args = args

        self.verbose = False
        self.debug = False
        if args.debug:
            if args.debug > 1:
                self.verbose = True
            if args.debug != 0:
                self.log.info('Debug mode set, not-saving any changes.')
                self.debug = True

    def run(self):
        """ The initial base for running the command.

        This needs to have a standardized format, argument list, and return
        either True or False depending on whether the command was successful or
        not.

        Returns:
            True if the command succeeded, otherwise False.
        """

        # Since this is just the base, it should always return True (because
        # you shouldn't fail at doing nothing).
        return True
